/*
    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "md5.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
#endif



#define VER         "0.1.1"
#define BUFFSZ      2048
#define HOST        "nwmaster.bioware.com"
#define PORT        5121
#define TIMEOUT     3



int build_nwn_pck(u_char *buff, u_char *user, u_char *pass);
void do_md5(u_char *out, u_char *in, int len);
void byte2hex(u_char *out, u_char *in, int len);
void create_rand(u_char *data, int len);
int timeout(int sock);
u_int resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    int     sd,
            len;
    u_char  buff[BUFFSZ],
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    fputs("\n"
        "Neverwinter Nights account checker "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <username> <password>\n"
            "\n", argv[0]);
        exit(1);
    }

    fputs("- build packet\n", stdout);
    len = build_nwn_pck(buff, argv[1], argv[2]);

    fputs("- resolv " HOST "...\n", stdout);
    peer.sin_addr.s_addr = resolv(HOST);
    peer.sin_port        = htons(PORT);
    peer.sin_family      = AF_INET;
    printf("- send packet to %s : %hu\n",
        inet_ntoa(peer.sin_addr), PORT);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(peer));
    if(timeout(sd) < 0) {
        fputs("\nError: socket timeout, no reply received\n\n", stdout);
        exit(1);
    }
    len = recvfrom(sd, buff, BUFFSZ, 0, NULL, NULL);
    if(len < 0) std_err();
    close(sd);

    if(memcmp(buff, "BMPR", 4)) {
        printf("\nError: wrong server reply (%.4s)\n\n", buff);
        exit(1);
    }
    p = buff + 4;
    p += *(u_short *)p + 2;
    if(!*(u_short *)p) {
        fputs("\n  Your account IS valid!!!\n\n", stdout);
    } else {
        fputs("\n  Login failed: your account doesn't seem valid\n\n", stdout);
    }

    return(0);
}



int build_nwn_pck(u_char *buff, u_char *user, u_char *pass) {
    int     len;
    u_char  tmp[64],
            *p;

    do_md5(tmp, pass, strlen(pass));
    create_rand(tmp + 32, 32);

#define ADDMEM(x,y) memcpy(p, x, y);    p += y;
#define ADD16(x)    *(u_short *)p = x;  p += 2;

    p = buff;
    ADDMEM("BMPA", 4);
    ADDMEM("\x00\x14", 2);
    ADD16(32);
    ADDMEM(tmp + 32, 32);
    len = strlen(user);
    ADD16(len);
    ADDMEM(user, len);
    ADD16(32);
    do_md5(p, tmp, 64);
    p += 32;
    ADDMEM("\x00\x00\x57\x02", 4);

#undef ADDMEM
#undef ADD16
    return(p - buff);
}



void do_md5(u_char *out, u_char *in, int len) {
    md5_context md5t;
    u_char      md5h[16];

    md5_starts(&md5t);
    md5_update(&md5t, in, len);
    md5_finish(&md5t, md5h);
    byte2hex(out, md5h, 16);
}



void byte2hex(u_char *out, u_char *in, int len) {
    const static u_char hex[] = "0123456789abcdef";

    while(len--) {
        *out++ = hex[*in >> 4];
        *out++ = hex[*in & 15];
        in++;
    }
}



void create_rand(u_char *data, int len) {
    u_int   rnd = time(NULL);

    while(len--) {
        rnd = (rnd * 0x343FD) + 0x269EC3;
        if(!(rnd & 0xff)) rnd++;    // only bytes different than 0x00!
        *data++ = rnd;
    }
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int  host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


